// The MIT License

// Copyright (c) 2010 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Author: Richard Uhler ruhler@mit.edu

import GetPut::*;
import FIFO::*;

import DDR2Types::*;


// Module adapts a non-bursting DDR2Client to a 4-burst DRR2Client.
// Non-bursting means for each command there is one 64 bit data word read or
// written. 4-burst means for each command there are four 64 bit data words
// read or written.
module mkDDR2FakeBurst#(DDR2Client nb) (DDR2Client);

    FIFO#(DDR2ReqData) wdata <- mkFIFO();
    FIFO#(DDR2Data) rdata <- mkFIFO();

    Reg#(Bit#(2)) wcount <- mkReg(0);
    Reg#(Bit#(2)) rcount <- mkReg(0);

    rule write (True);
        if (wcount == 0) begin
            let x <- nb.request.data.get();
            wdata.enq(x);
        end else begin
            wdata.enq(DDR2ReqData { data: ?, be: 8'h00 });
        end
        wcount <= wcount+1;
    endrule

    rule read (True);
        if (rcount == 0) begin
            nb.response.put(rdata.first());
        end
        rcount <= rcount+1;
        rdata.deq();
    endrule

    interface DDR2RequestClient request;
        interface Get command = nb.request.command;
        interface Get data = toGet(wdata);
    endinterface

    interface Put response = toPut(rdata);

endmodule

